from django.conf import settings
from django.core.urlresolvers import (
    get_resolver,
    resolve,
    reverse,
    RegexURLPattern,
    RegexURLResolver,
)
from django.template.response import SimpleTemplateResponse
from django.test import RequestFactory
from django.views.generic import TemplateView
import os
import os.path
import sys


class Bakeable:
    """A CBV that can be baked. This is pretty abstract."""

    def bake(self, output_dir=None, paths=None, verbosity=0, stdout=sys.stdout):
        """
        Bake one or more URLs via this view into output_dir
        (defaults to settings.FBO_BUILD_DIR). If paths is
        not None, then provides a list of URL paths to
        bake, otherwise bake all paths possible for this
        view (generally must be implemented by the view).
        """

        if output_dir is None:
            output_dir = settings.FBO_BUILD_DIR
        if paths is None:
            if verbosity > 1:
                stdout.write(" > fetching paths (in %s)\n" % str(self))
            paths = self.get_paths()

        factory = RequestFactory()
        for path in paths:
            if verbosity > 2:
                stdout.write(" * %s\n" % str(path))
            filename = self.get_filename(path)
            out_fname = os.path.join(output_dir, filename)
            os.makedirs(
                os.path.dirname(out_fname),
                exist_ok=True,
            )
            with open(out_fname, 'wb') as f:
                match = resolve(path)
                request = factory.get(path)
                response = match.func(
                    request,
                    *match.args,
                    **match.kwargs
                )
                if isinstance(response, SimpleTemplateResponse):
                    response.render()
                if response.status_code // 100 == 2:
                    if response.streaming:
                        # FIXME: do something with this.
                        f.write(
                            "Unhandled: streaming responses.".encode('utf-8'),
                        )
                        pass
                    else:
                        # FIXME: should inject a meta header into HTML
                        # so that .charset is preserved. There may be
                        # other headers on the response that are worth
                        # preserving in similar ways.
                        f.write(response.content)
                else:
                    # FIXME redirects we could trap and write out
                    # either HTML files with meta refresh, or suitable
                    # configuration for Apache, nginx &c. (Or both.)
                    f.write(
                        (
                            "Unhandled status code %i." % response.status_code
                        ).encode('utf-8'),
                    )

    def get_filename(self, path):
        # FIXME: won't work with non-Unixoid file names.
        if path.startswith('/'):
            # Which should be always!
            path = path[1:]
        elif path!='':
            raise ValueError("Path should start with '/'.")
        components = path.split('/')
        if components[-1] == '':
            components[-1] = 'index'
        if '.' not in components[-1]:
            components[-1] += '.html'

        return '/'.join(components)

    def get_paths(self):
        """
        Returns a sequence of paths, ie netloc-relative URLs, each of
        which will be turned into a ResolverMatch via resolve(), and
        then rendered 'as normal' (using a mock HttpRequest object
        via django.test.client.RequestFactory).

        By default, figures out the URL using reverse().
        This will only work for singleton views.
        """

        return tuple(
            get_resolver().reverse(self)
        )


class BakeableTemplateView(Bakeable, TemplateView):
    # This one just works, so make it simple to use.
    pass


def bake(output_dir=None, resolver=None, verbosity=0, stdout=sys.stdout):
    if resolver is None:
        resolver = get_resolver()
    if verbosity > 1:
        stdout.write("Baking %s\n" % str(resolver))
    if isinstance(resolver, RegexURLResolver):
        for up in resolver.url_patterns:
            bake(output_dir, up, verbosity, stdout)
    elif isinstance(resolver, RegexURLPattern):
        view = resolver.callback
        # `view` is a callable, but it may also be
        # an instance of a CBV, which is all we support.
        # If it's a subclass of our Bakeable then
        # we can bake it, because it provides the support
        # to do so.
        if hasattr(view, 'view_class'):
            view_class = view.view_class
            view_instance = view_class(**view.view_initkwargs)
            if isinstance(view_instance, Bakeable):
                if resolver.regex.groups == 0:
                    # This is unique, so ensure this URL gets baked.
                    # Otherwise (eg) homepages are really hard to
                    # get right. You must provide a name for this
                    # to work; otherwise we SILENTLY do nothing, which
                    # isn't friendly.
                    #
                    # Note this may mean that we write something out
                    # twice. This really shouldn't cause problems,
                    # although it probably will at some point.
                    if resolver.name is not None:
                        view_instance.bake(
                            output_dir,
                            [reverse(resolver.name)],
                            verbosity,
                            stdout,
                        )
                else:
                    view_instance.bake(
                        output_dir,
                        None,
                        verbosity,
                        stdout,
                    )
            else:
                if verbosity > 0:
                    stdout.write(" ! skipping %s\n" % str(view))



import fnmatch
from django.db.models.query_utils import Q as ORM_Q


class Q(ORM_Q):
    """Encapsulates filters as objects that can then be combined logically."""
    # Connection types
    AND = all
    OR = any
    default = AND

    # Most of the mechanics (the tree, operations, cloning) is taken
    # from Django's ORM Q object. This means we need to remove a couple
    # of methods from the public API.
    #
    # The exception we use isn't strictly accurate, but gives
    # equivalent behaviour to deleting the methods entirely (which we
    # can't do because they come via inheritance).

    def resolve_expression(self, *args, **kwargs):
        raise AttributeError(
            "'django_FBO.Q' object has no attribute 'resolve_expression'",
        )

    def refs_aggregate(self, *args, **kwargs):
        raise AttributeError(
            "'django_FBO.Q' object has no attribute 'refs_aggregate'",
        )

    # And we don't contribute to a SQL query; the tree acts as a
    # functor, applied as a filter on the incoming stream of all
    # files.
    
    def __call__(self, _file):
        def _check_one(child):
            # child is a node (one of us) or a tuple
            if isinstance(child, Q):
                return child(_file)
            else:
                _filter, _filter_val = child
                if '__' in _filter:
                    _field, _operator = _filter.split('__', 2)
                else:
                    _field, _operator = _filter, 'equals'

                _field_val = getattr(_file, _field)
                op = Operators.get(_operator)
                if op is None:
                    raise ValueError(
                        "No such operator '%s' in filter '%s'" % (
                            _operator,
                            _filter,
                        )
                    )

                _op_result = op(_field, _field_val, _filter_val)
                #print(
                #    "Filter %s.%s__%s=%s: %s -> %s" % (
                #        _file,
                #        _field,
                #        _operator,
                #        _filter_val,
                #        _field_val,
                #        _op_result,
                #    )
                #)
                return _op_result

        _check_results = [
            _check_one(c) for c in self.children
        ]
        if self.negated:
            return not self.connector(_check_results)
        else:
            return self.connector(_check_results)


def globerator(field, field_val, filter_val):
    return fnmatch.fnmatch(field_val, filter_val)

def equals(field, field_val, filter_val):
    return field_val == filter_val

def lte(field, field_val, filter_val):
    return field_val <= filter_val

def gte(field, field_val, filter_val):
    return field_val >= filter_val

def lt(field, field_val, filter_val):
    return field_val < filter_val

def gt(field, field_val, filter_val):
    return field_val > filter_val

def contains(field, field_val, filter_val):
    return filter_val in field_val

def startswith(field, field_val, filter_val):
    return str(field_val).startswith(filter_val)

def endswith(field, field_val, filter_val):
    return str(field_val).endswith(filter_val)

def in_operator(field, field_val, filter_val):
    return field_val in filter_val

Operators = {
    'glob': globerator,
    'equals': equals,
    'lte': lte,
    'gte': gte,
    'lt': lt,
    'gt': gt,
    'contains': contains,
    'startswith': startswith,
    'endswith': endswith,
    'in': in_operator,
}



"""
WSGI config for jetkittens project.

It exposes the WSGI callable as a module-level variable named ``application``.
"""

import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "@@SITEDIR@@.settings")

application = get_wsgi_application()



from django.core.exceptions import (
    ObjectDoesNotExist,
    MultipleObjectsReturned,
)
from django.db.models.fields import TextField
import json
import yaml


class Options:
    verbose_name = None
    verbose_name_plural = None
    app_label = None
    model_name = None

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_field(self, field_name, many_to_many=None):
        # many_to_many is bc from pre-1.9, so we can
        # ignore it since we really only support 1.10
        # fully (even though it's not out yet ;-).
        return TextField(name=field_name)


class FileObjectMeta(type):

    def __new__(cls, name, bases, namespace, **kwds):
        new_class = type.__new__(cls, name, bases, namespace, **kwds)
        # FIXME we should use a Django-like `cls.Meta` pattern here
        # to allow overriding.
        # Possibly the way we pass them through into Options isn't
        # great, either.
        #
        # FIXME: these really should be generated :-)
        verbose_name = 'thing'
        verbose_name_plural = 'things'
        model_name = 'who'
        app_label = 'whatever'

        new_class._meta = Options(
            verbose_name=verbose_name,
            verbose_name_plural=verbose_name_plural,
            app_label=app_label,
            model_name=model_name,
        )
        return new_class


class FileObject(metaclass=FileObjectMeta):
    MetadataInFileHead = True
    DoesNotExist = ObjectDoesNotExist
    MultipleObjectsReturned = MultipleObjectsReturned

    def __init__(
        self,
        storage,
        metadata_location,
        name,
        slug_suffices=None,
        slug_strip_index=None,
    ):
        self.storage = storage
        self.metadata_location = metadata_location
        self.name = name
        self.path = storage.path(name=name)
        self.slug_suffices = slug_suffices
        self.slug_strip_index = slug_strip_index
        self._metadata = None

    @property
    def slug(self):
        """
        All the Django CBVs work in terms of slug, which is
        a common enough pattern that it's useful to just map
        that to name. You may want to override this if the
        slug is only part of your name (for instance see the
        blog module).

        slug_suffices is passed down from the FBO, and is a
        list of suffices that can be stripped. It's down to
        you to ensure there aren't collisions, or your FBO
        .get() won't always return the one you expect.
        """

        slug = self.name
        try:
            for suffix in self.slug_suffices:
                if slug.endswith(suffix):
                    slug = slug[:-len(suffix)]
                    break
        except TypeError:
            pass

        if self.slug_strip_index:
            # We want to transform .../index -> .../
            # so find the last component and convert
            components = slug.split('/')
            if components[-1] == 'index':
                components[-1] = ''
                slug = '/'.join(components)
            if slug == '/':
                # this is a special case, because we want
                # root indexes to work with a url pattern of:
                # r'^(?P<slug>.*)$' which will be empty at
                # the root.
                slug = ''
        return slug

    @property
    def metadata(self):
        if self._metadata is None:
            self._metadata = self._load_metadata()
        return self._metadata

    def _load_content(self):
        with self.storage.open(self.name) as _file:
            return _file.read().decode('utf-8')

    def _load_metadata(self):
        self.content = self._load_content()
        if self.metadata_location == FileObject.MetadataInFileHead:
            if self.content.startswith('{\n'):
                # JSON!
                end = self.content.find('}\n')
                if end!=-1:
                    blob = self.content[:end+2]
                    self.content = self.content[end+2:]
                    data = json.loads(blob)
                    return data
            elif self.content.startswith('---\n'):
                # YAML!
                # Magic numbers: 4 is skipping the intro ---\n,
                # 8 is skipping both intro and outro ---\n.
                end = self.content[4:].find('---\n')
                if end!=-1:
                    blob = self.content[4:end+4]
                    self.content = self.content[end+8:]
                    data = yaml.load(blob)
                    return data
            else:
                # Implicit YAML if ':' before \n\n
                colon_idx = self.content.find(':')
                sep_idx = self.content.find('\n\n')
                # If sep_idx is -1 or both are, first leg won't pass
                if colon_idx < sep_idx and colon_idx != -1:
                    # YAML!
                    blob = self.content[:sep_idx]
                    self.content = self.content[sep_idx+2:]
                    data = yaml.load(blob)
                    return data
        return {}

    def __getattr__(self, key):
        # if we are asked for content before any metadata,
        # need to load it
        if key == 'content':
            _ = self.metadata
            return self.content
        return self.metadata.get(key, None)

    def __eq__(self, other):
        if other is None or not isinstance(other, FileObject):
            return False
        return self.path == other.path

    def __str__(self):
        return self.name



"""
Django settings for @@SITENAME@@ project.

For more information on this file, see
https://docs.djangoproject.com/en/1.10/topics/settings/
"""

import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

DEBUG = os.environ.get('DJANGO_DEBUG', True)
if DEBUG:
    SECRET_KEY = '@@SECRET_KEY@@'
else:
    SECRET_KEY = os.environ['SECRET_KEY']
ALLOWED_HOSTS = os.environ.get('ALLOWED_HOSTS', '').split(';')

# Application definition

INSTALLED_APPS = [
    'django.contrib.staticfiles',
    'django_cape',
    'django_FBO',
    'markdown_deux',
]

MIDDLEWARE_CLASSES = [
    'django.middleware.security.SecurityMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = '@@SITEDIR@@.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
            ],
        },
    },
]

WSGI_APPLICATION = '@@SITEDIR@@.wsgi.application'

DATABASES = {}
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True
LANGUAGE_CODE = 'en-gb'

# Markdown settings
MARKDOWN_DEUX_STYLES = {
    'default': {
        'extras': {
            'smarty-pants': True,
            'code-friendly': None,
            'wiki-tables': True,
        },
        'safe_mode': 'escape',
    },
}

# Static files (CSS, JavaScript, Images)
STATIC_URL = '/static/'
STATICFILES_DIRS = [
    os.path.join(BASE_DIR, 'static')
]
STATIC_ROOT = os.path.join(BASE_DIR, 'build', 'static')

# Media files
MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')

# By default, strip .md from slugs (so something.md -> something)
#
# Note that this means that for Pages, having pages/about will
# response to /about (given the default URL configuration), which will
# bake to .../about.html and will require something like Apache's
# Options +MultiViews to serve properly. If you can't set things up
# that way, then use pages/about/index.md and you'll get /about/ and
# .../about/index.html, which only requires Options +Indexes or
# similar.
#
# The default is None, ie don't strip any slug suffices. (index will
# still be stripped though.)
FBO_DEFAULT_SLUG_SUFFICES = [ '.md' ]
# Where do we bake FBO to?
FBO_BUILD_DIR = os.path.join(BASE_DIR, 'build')

@@MODULE_SETTINGS@@



from django.conf import settings
from django.core.management.base import BaseCommand

from ...baking import bake


class Command(BaseCommand):
    help = """Bake all FBO-based views."""

    def add_arguments(self, parser):
        parser.add_argument(
            'outdir',
            nargs='?',
            help='directory to build to (defaults to %s)' % settings.FBO_BUILD_DIR,
        )

    def handle(self, *args, **options):
        bake(
            output_dir=options['outdir'],
            verbosity=options['verbosity'],
            stdout=self.stdout,
        )



# django_FBO newsite script (installed as a console_script entry point)

import argparse
from django.utils.crypto import get_random_string
from os import mkdir
import os.path
import sys

# Default context for all renders.
_context = {}

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

def mkfile(template, prefix=''):
    """
    Create a single file, named after the template, in the directory
    `prefix` (use the empty string to avoid a subdirectory).

    Templates are in `site_templates`.

    Templates are VERY SIMPLISTIC, they just replace variables of the
    form @@VARNAME@@ from _context.  This is done in order,
    non-recursively; we aren't parsing anything "properly".

    Missing template creates an empty file.

    """

    with open(os.path.join(prefix, template), 'wb') as fp:
        try:
            with open(
                os.path.join(THIS_DIR, 'site_templates', template),
                'rb',
            ) as tfp:
                template = tfp.read().decode('utf-8')
                for varname, value in _context.items():
                    template = template.replace('@@%s@@' % varname, value)
            fp.write(template.encode('utf-8'))
        except FileNotFoundError:
            # No template, ie empty file (eg __init__.py)
            pass
    

def newsite(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog='django-FBO-newsite',
        description='Start a new Django FBO site.',
    )
    parser.add_argument(
        'sitedir', nargs=1,
        help='subdirectory for site configuration',
    )
    parser.add_argument(
        'sitename', nargs=1,
        help='name of new site',
    )
    parser.add_argument(
        '--modules', dest='modules', action='store',
        help='modules to enable, comma separated',
    )

    args = parser.parse_args(args)
    if args.modules is None:
        modules = []
    else:
        modules = args.modules.split(',')

    sitedir = args.sitedir[0]
    _context['SITENAME'] = args.sitename[0]
    _context['SITEDIR'] = sitedir
    chars = 'abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)'
    _context['SECRET_KEY'] = get_random_string(50, chars)
    URLS = ''
    MODULE_SETTINGS = ''
    URL_IMPORTS = ''

    for module in modules:
        if module == 'pages':
            URL_IMPORTS += """
from django_FBO.modules.pages import PageView
"""
            URLS += """
    url(r'^(?P<slug>.*)$', PageView.as_view(), name='page'),
"""
        if module == 'blog':
            URLS += """
    url(r'^blog/', include('django_FBO.modules.blog')),
"""
            MODULE_SETTINGS += """
# FBO blog settings
FBO_BLOG_TITLE = u'My new blog'
FBO_BLOG_SUBTITLE = u'Some thoughts from someone on the internet'
FBO_BLOG_COPYRIGHT = u'(c) Copyright me, probably'
FBO_BUILD_DIR = os.path.join(BASE_DIR, 'build')
"""

    _context['URLS'] = URLS
    _context['URL_IMPORTS'] = URL_IMPORTS
    _context['MODULE_SETTINGS'] = MODULE_SETTINGS

    # We make a normal Django structure, with some strong decisions.
    # 1. Settings follow 12 Factor, broadly.
    # 2. Use CAPE, via django_cape.
    # 3. Set up very simple templates and styling.
    mkfile('manage.py')
    # FIXME: +x manage.py (within umask)
    mkfile('requirements.txt')
    mkdir(sitedir)
    mkfile('__init__.py', sitedir)
    mkfile('settings.py', sitedir)
    mkfile('urls.py', sitedir)
    mkfile('wsgi.py', sitedir)

    # From here on is broadly 'the theme'.
    mkdir('media')
    mkdir('static')
    mkdir(os.path.join('static', 'css'))
    mkfile(os.path.join('static', 'css', 'style.css'))

    mkdir('templates')
    mkfile(os.path.join('templates', 'home.html'))
    mkfile(os.path.join('templates', 'base.html'))

    for module in modules:
        if module == 'pages':
            mkfile(os.path.join('templates', 'page.html'))
        elif module == 'blog':
            mkdir(os.path.join('templates', 'blog'))
            mkfile(os.path.join('templates', 'blog', 'index.html'))
            mkfile(os.path.join('templates', 'blog', 'post.html'))
            mkfile(os.path.join('templates', 'blog', 'draft.html'))
            mkfile(os.path.join('templates', 'blog', '_pagination.html'))

if __name__ == "__main__":
    newsite()



"""@@SITENAME@@ URL Configuration"""
from django.conf import settings
from django.conf.urls import include, url
from django.conf.urls.static import static
from django.views.generic import TemplateView
@@URL_IMPORTS@@

urlpatterns = [
    url(r'^$', TemplateView.as_view(template_name='home.html'), name='home'),
@@URLS@@
]

if settings.DEBUG:
    urlpatterns += static(
        settings.MEDIA_URL, document_root=settings.MEDIA_ROOT,
    )



"""Binary module for django_FBO.

Drop this into your urlconf, perhaps as:

url(r'^binary/(?P<slug>.*)$', ObjectView.as_view(), name='binary'),

and drop things into {{ BASE_DIR }}/binaries. However that's unusual,
because you can use MEDIA_ROOT and MEDIA_URL for that. More likely is
that you want pages and binary objects interspersed, in which case you
want to use the interspersed module.

This isn't bootstrappable via django-fbo-newsite, because it's
somewhat niche. I use it to migrate existing sites in, but I'd be
unlikely to use it for brand new ones. Either way, there's usually
going to be some jockeying to make things work.
"""

from django.conf import settings
from django.core.urlresolvers import reverse
import os.path

from .. import FileObject
from .pages import Page, PageView


class BinaryFileObject(FileObject):

    def get_absolute_url(self):
        # [1] to get the ext, then strip a leading '.'
        ext = os.path.splitext(self.name)[1].lstrip('.')
        return reverse(
            ext,
            kwargs={
                'slug': self.slug,
            }
        )

    def _load_content(self):
        with self.storage.open(self.name) as _file:
            return _file.read()


class BinaryFBO(Page):
    metadata = None
    model = BinaryFileObject
    path = os.path.join(settings.BASE_DIR, 'binaries')


class BinaryView(PageView):
    queryset = BinaryFBO()



"""Interspersed binary & pages module for django_FBO.

Put binary and pages in {{ BASE_DIR }}/pages, and:

urlpatterns += get_interspersed_urls(
    exts=[ 'jpg', 'png', 'pdf' ],
)

which will give you both your page and binary URLs. Alternatively,
if you want to prefix things, you can use:

urlpatterns = [
    url(r'^prefix/', include('django_FBO.modules.interspersed')),
]

and set the following in settings:

FBO_INTERSPERSED_EXTS = [ 'jpg', 'png', 'pdf' ]

(The default list is empty.)
"""

from django.conf import settings
from django.conf.urls import url
from django.http import HttpResponse
from functools import reduce
import mimetypes
import os.path

from .. import Q
from .pages import PageView
from .binary import BinaryFBO, BinaryView


class InterspersedBinaryFBO(BinaryFBO):
    path = os.path.join(settings.BASE_DIR, 'pages')


class InterspersedBinaryView(BinaryView):
    queryset = InterspersedBinaryFBO()
    ext = None

    def get_queryset(self):
        queryset = super().get_queryset()
        if self.ext is None:
            # This is weird; there isn't a good reason to do this,
            # and plenty of reasons not to.
            return self.queryset
        else:
            return self.queryset.filter(name__glob='*.%s' % self.ext)

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        mime_type, encoding = mimetypes.guess_type(
            self.object.path,
            strict=False,
        )
        if mime_type is None:
            mime_type = 'application/octet-stream'
        resp = HttpResponse(
            content=self.object.content,
            content_type=mime_type,
        )
        if encoding is not None:
            resp['Content-Encoding'] = encoding
        return resp


class InterspersedPageView(PageView):
    pass


def get_interspersed_urls(exts, binaryview_kwargs=None, pageview_kwargs=None):
    # Note: you can prefix your URLs via existing urlpatterns mechanisms.

    # FIXME: this could be more efficient by making a single FBO with
    # pageview_excludes as a filter, then enumerating it, and passing
    # that to all the binary views. (It should be possible to avoid
    # walking the tree so much this way. At the moment it'll be done
    # once per request, which is tedious.)
    urls = []
    pageview_excludes = []
    if binaryview_kwargs is None:
        binaryview_kwargs = {}
    for ext in exts:
        urls.append(
            url(
                r'^(?P<slug>.*\.%(ext)s)$' % {
                    'ext': ext,
                },
                InterspersedBinaryView.as_view(
                    ext=ext,
                    **binaryview_kwargs
                ),
                name=ext,
            ),
        )
        pageview_excludes.append(
            Q(name__glob='*.%s' % ext),
        )
    if pageview_kwargs is None:
        pageview_kwargs = {}
    if 'queryset' not in pageview_kwargs:
        pageview_kwargs['queryset'] = InterspersedPageView.queryset
    pageview_kwargs['queryset'] = pageview_kwargs['queryset'].exclude(
        reduce(
            lambda x, y: x | y,
            pageview_excludes,
        )
    )
    urls.append(
        url(
            r'^(?P<slug>.*)$',
            InterspersedPageView.as_view(**pageview_kwargs),
            name='page',
        )
    )
    return urls


urlpatterns = get_interspersed_urls(
    getattr(settings, 'FBO_INTERSPERSED_EXTS', []),
)



import os.path
from datetime import datetime
from django.conf import settings
from django.conf.urls import url
from django.contrib.syndication.views import Feed as _Feed
from django.core.urlresolvers import reverse
from django.db.models.fields import TextField
from django.utils import timezone
from django.utils.feedgenerator import Atom1Feed
from django.views.generic import (
    DetailView as _DetailView,
    ArchiveIndexView as _ArchiveIndexView,
    YearArchiveView as _YearArchiveView,
    MonthArchiveView as _MonthArchiveView,
    DayArchiveView as _DayArchiveView,
    DateDetailView as _DateDetailView,
    ListView,
)
from markdown_deux import markdown
from .. import FBO, FileObject, Q, Bakeable


def get_drafts_prefix():
    # Note this must function in a regular expression as well.
    return getattr(settings, 'FBO_BLOG_DRAFTS_PREFIX', 'drafts/')


class Author(object):
    # Currently just used for generating the feed, but
    # in theory could be populated automatically for a
    # BlogPostFile and so made more useful.
    #
    # Note that right now we don't actually populate the
    # url at all. This is laziness and not having needed
    # to as yet.
    def __init__(self, name, url=None):
        self.name = name
        self.url = url


class BlogPostFile(FileObject):

    def get_absolute_url(self):
        if self.slug.startswith(get_drafts_prefix()):
            return reverse(
                'blog-draft-detail',
                kwargs={
                    'slug': self.slug,
                },
            )
        else:
            return reverse(
                'blog-detail',
                kwargs={
                    'year': '%04d' % self.date.year,
                    'month': '%02d' % self.date.month,
                    'day': '%02d' % self.date.day,
                    'slug': self.slug,
                }
            )

    @property
    def slug(self):
        """
        If we can split the slug into YEAR/MONTH/DAY/SLUG,
        do so; otherwise the slug remains unchanged.
        """

        try:
            slug = super().slug
        except:
            raise KeyError('slug')
        try:
            _, _, _, slug = slug.split('/', 4)
            return slug
        except:
            return slug

    @property
    def date(self):
        # figure out the date from the name, else delegate
        # to published
        try:
            year, month, day, _ = self.name.split('/', 4)
            return timezone.now().replace(
                year=int(year),
                month=int(month),
                day=int(day),
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            )
        except:
            return self.published

    @property
    def published(self):
        if 'published' in self.metadata:
            return self.metadata['published']
        else:
            return self.storage.get_created_time(self.name)


class BlogPost(FBO):
    path = os.path.join(settings.BASE_DIR, 'posts')
    _filters = [
        ~Q(name__glob='*~'),
    ]
    _order_by = [ 'date' ]
    metadata = BlogPostFile.MetadataInFileHead
    model = BlogPostFile
    slug_suffices = getattr(settings, 'FBO_DEFAULT_SLUG_SUFFICES', None)
    slug_strip_index = True

    def exclude_drafts(self):
        return self.exclude(
            Q(status='draft') |
            Q(name__startswith=get_drafts_prefix()),
        )

    def find_next(self, obj):
        _iter = iter(self.exclude_drafts())
        try:
            for candidate in _iter:
                if candidate == obj:
                    return next(_iter)
        except StopIteration:
            return None

    def find_previous(self, obj):
        _iter = iter(self.exclude_drafts())
        previous = None
        try:
            for candidate in _iter:
                if candidate == obj:
                    return previous
                else:
                    previous = candidate
        except StopIteration:
            return None


class BakeableBlogMixin(Bakeable):
    template_name = 'blog/index.html'
    queryset = BlogPost().exclude_drafts()
    date_field = 'date'
    uses_datetime_field = True
    paginate_by = 10
    make_object_list = True

    def get_url_name(self):
        return self.url_name

    def _paginate_for_baking(self, qs, date):
        page_size = self.get_paginate_by(qs)
        if page_size:
            self.kwargs = { 'page': 1 }
            paginator, page, queryset, is_paginated = self.paginate_queryset(qs, page_size)
            for page in range(paginator.num_pages):
                yield self.date_to_url(date, page+1)
        else:
            yield self.date_to_url(date)

    def get_date_to_url_kwargs(self, date):
        raise NotImplementedError

    def date_to_url(self, date, page=None):
        kwargs = self.get_date_to_url_kwargs(date)
        if page is None or page == 1:
            return reverse(
                self.get_url_name(),
                kwargs=kwargs,
            )
        else:
            kwargs['page'] = page
            return reverse(
                self.get_url_name() + '-paginated',
                kwargs=kwargs,
            )


class ArchiveIndexView(BakeableBlogMixin, _ArchiveIndexView):
    url_name = 'blog-index'

    def get_paths(self):
        return self._paginate_for_baking(
            self.get_queryset(),
            timezone.now(),
        )

    def get_date_to_url_kwargs(self, date):
        return {}


class BakeableBlogDateMixin(BakeableBlogMixin):
    month_format = '%m'

    def get_date_list_period_for_baking(self):
        period = self.get_url_name()
        if period.startswith('blog-'):
            period = period[5:]
        return period

    def get_paths(self):
        # Marvel at how much duplication there is between this and the
        # CBV date system! All the fun of millions of classes, without
        # the extension hooks you actually need.
        kind = self.get_date_list_period_for_baking()

        for date in self.get_queryset().datetimes(
            field_name=self.get_date_field(),
            kind=kind,
            ordering='ASC',
        ):
            # For each date, paginate!
            date_field = self.get_date_field()
            since = self._make_date_lookup_arg(date)
            # _get_next_year, _get_next_month, _get_next_day
            until = self._make_date_lookup_arg(
                getattr(self, '_get_next_' + kind)(date)
            )
            lookup_kwargs = {
                '%s__gte' % date_field: since,
                '%s__lt' % date_field: until,
            }
            qs = self.get_dated_queryset(**lookup_kwargs)
            for path in self._paginate_for_baking(qs, date):
                yield path


class YearArchiveView(BakeableBlogDateMixin, _YearArchiveView):
    url_name = 'blog-year'

    def get_date_to_url_kwargs(self, date):
        return {
            'year': '%04d' % date.year,
        }


class MonthArchiveView(BakeableBlogDateMixin, _MonthArchiveView):
    url_name = 'blog-month'

    def get_date_to_url_kwargs(self, date):
        return {
            'year': '%04d' % date.year,
            'month': '%02d' % date.month,
        }


class DayArchiveView(BakeableBlogDateMixin, _DayArchiveView):
    url_name = 'blog-day'

    def get_date_to_url_kwargs(self, date):
        return {
            'year': '%04d' % date.year,
            'month': '%02d' % date.month,
            'day': '%02d' % date.day,
        }


class DateDetailView(Bakeable, _DateDetailView):
    template_name = 'blog/post.html'
    queryset = BlogPost()
    date_field = 'date'
    uses_datetime_field = True
    month_format = '%m'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['next_object'] = self.queryset.find_next(self.object)
        context['previous_object'] = self.queryset.find_previous(self.object)
        return context

    def get_paths(self):
        # Remember that get_absolute_url() doesn't return
        # an absolute URL, just netloc-relative.
        return [
            p.get_absolute_url()
            for p in self.queryset.all()
        ]


class DraftsList(ListView):
    queryset = BlogPost().filter(
        Q(name__startswith=get_drafts_prefix())
    )
    paginate_by = None

    def get_template_names(self):
        return [ 'blog/drafts_index.html', 'blog/index.html' ]


class DraftDetailView(Bakeable, _DetailView):
    template_name = 'blog/draft.html'
    queryset = BlogPost().filter(
        name__startswith=get_drafts_prefix(),
    )

    def get_paths(self):
        # Remember that get_absolute_url() doesn't return
        # an absolute URL, just netloc-relative.
        return [
            p.get_absolute_url()
            for p in self.queryset.all()
        ]


class BetterAtomFeed(Atom1Feed):
    def add_item_elements(self, handler, item):
        super(BetterAtomFeed, self).add_item_elements(handler, item)
        authors = item.get('authors', [])
        for author in authors:
            if author is None:
                continue
            handler.startElement("author", {})
            handler.addQuickElement("name", author.name)
            if author.url is not None:
                handler.addQuickElement("uri", author.url)
            handler.endElement("author")


class Feed(_Feed):
    feed_type = BetterAtomFeed

    def __init__(self, paginate_by):
        self.paginate_by = paginate_by

    def items(self):
        return self.queryset[:self.paginate_by]

    def item_title(self, item):
        return item.title

    def item_pubdate(self, item):
        return item.date

    def item_updateddate(self, item):
        return item.date

    def item_description(self, item):
        return markdown(item.content)

    def item_extra_kwargs(self, item):
        extra = {}
        if hasattr(item, 'author'):
            # FIXME: pull url if it exists
            author_obj = Author(item.author, None)
            extra['authors'] = [ author_obj ]
        return extra


class FeedMixin(object):
    """
    Mixin to a ListView to output an Atom feed instead of HTML.
    """

    # empty Atom feeds are empty, not 404 (ListView behaviour)
    allow_empty = True
    # conventionally, feeds have 10 items
    paginate_by = 10

    # Defaults for feed-level properties.
    feed_title = None
    feed_subtitle = None
    feed_link = None
    feed_copyright = None

    # Feed class itself
    feed_class = Feed

    def get_title(self):
        return self.feed_title

    def get_subtitle(self):
        return self.feed_subtitle

    def get_link(self):
        return self.feed_link

    def get_copyright(self):
        return self.feed_copyright

    def render_to_response(self, context):
        feedview = self.feed_class(self.paginate_by)
        feedview.title = self.get_title()
        feedview.subtitle = self.get_subtitle()
        feedview.link = self.get_link()
        feedview.queryset = self.get_queryset()
        feedview.feed_copyright = self.get_copyright()
        feedview.item_copyright = self.get_copyright()
        return feedview(self.request, *self.args, **self.kwargs)


class BlogFeed(FeedMixin, ArchiveIndexView):
    blog_index_url_name = 'blog-index'

    def get_link(self):
        return reverse('blog-index')



# Just include this directly, unless you want to customise
# any of the view defaults.
urlpatterns = [
    # First our indexes. Each comes twice, the second URL dealing with
    # pagination. We have: everything, year, month and day.
    url(r'^$', ArchiveIndexView.as_view(), name='blog-index'),
    url(r'^p(?P<page>[0-9]+)/$', ArchiveIndexView.as_view(), name='blog-index-paginated'),
    url(r'^(?P<year>[0-9]{4})/$', YearArchiveView.as_view(), name='blog-year'),
    url(r'^(?P<year>[0-9]{4})/p(?P<page>[0-9]+)/$', YearArchiveView.as_view(), name='blog-year-paginated'),
    url(r'^(?P<year>[0-9]{4})/(?P<month>[0-9]{2})/$', MonthArchiveView.as_view(), name='blog-month'),
    url(r'^(?P<year>[0-9]{4})/(?P<month>[0-9]{2})/p(?P<page>[0-9]+)/$', MonthArchiveView.as_view(), name='blog-month-paginated'),
    url(r'^(?P<year>[0-9]{4})/(?P<month>[0-9]{2})/(?P<day>[0-9]{2})/$', DayArchiveView.as_view(), name='blog-day'),
    url(r'^(?P<year>[0-9]{4})/(?P<month>[0-9]{2})/(?P<day>[0-9]{2})/p(?P<page>[0-9]+)/$', DayArchiveView.as_view(), name='blog-day-paginated'),
    # Then a date-based view of a single blog post.
    url(r'^(?P<year>[0-9]{4})/(?P<month>[0-9]{2})/(?P<day>[0-9]{2})/(?P<slug>.*)$', DateDetailView.as_view(), name='blog-detail'),
    # Now our Atom feed.
    url(r'^index.atom$', BlogFeed.as_view(
        feed_title = settings.FBO_BLOG_TITLE,
        feed_subtitle = settings.FBO_BLOG_SUBTITLE,
        feed_copyright = settings.FBO_BLOG_COPYRIGHT,
    ), name='blog-feed'),
    # Drafts list -- this doesn't get baked, so it won't appear in the
    # live site.
    url(r'^drafts/$', DraftsList.as_view()),
    # And finally our drafts. Note that we include the prefix directly
    # into the regular expression, so the prefix must not contain any
    # RE atoms that aren't literals.
    url(
        r'^(?P<slug>' + get_drafts_prefix() + '.*)$',
        DraftDetailView.as_view(),
        name='blog-draft-detail',
    ),
]



"""Pages module for django_FBO.

Looks for things in {{ BASE_DIR }}/pages, by default ignoring *~
backup files. Include in your urlconf as something like:

url(r'^(?P<slug>.*)$', Page.as_view(), name='page'),

See the interspersed module if you want to have (eg) images in the
same URL hierarchy as your pages (instead of putting them in /media/
or similar, which you can do using MEDIA_ROOT and MEDIA_URL).

"""

import os.path
from django.conf import settings
from django.core.urlresolvers import reverse
from django.views.generic import DetailView

from .. import FileObject, FBO, Q, Bakeable


class PageFile(FileObject):

    def get_absolute_url(self):
        return reverse(
            'page',
            kwargs={
                'slug': self.slug,
            }
        )


class Page(FBO):
    path = os.path.join(settings.BASE_DIR, 'pages')
    _filters = [
        ~Q(name__glob='*~'),
    ]
    metadata = FileObject.MetadataInFileHead
    model = PageFile
    slug_suffices = getattr(settings, 'FBO_DEFAULT_SLUG_SUFFICES', None)
    slug_strip_index = True


class PageView(Bakeable, DetailView):
    template_name = 'page.html'
    queryset = Page()
    slug = None

    def get_template_names(self):
        override =  self.get_object().metadata.get('template')
        if override is not None:
            return [ override ]
        else:
            return [ self.template_name ]

    def get_object(self, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()
        if self.slug is not None:
            return queryset.get(slug=self.slug)
        else:
            return super(PageView, self).get_object(queryset)

    def get_paths(self):
        # Remember that get_absolute_url() doesn't return
        # an absolute URL, just netloc-relative.
        return [
            p.get_absolute_url()
            for p in self.get_queryset().all()
        ]



from setuptools import setup

PACKAGE = 'django_FBO'
VERSION = '3.4.0'

setup(
    name=PACKAGE,
    version=VERSION,
    description="Access file-backed objects in something resembling the ORM.",
    packages=[
        'django_FBO',
        'django_FBO.modules',
        'django_FBO.management',
        'django_FBO.management.commands',
    ],
    package_data={
        'django_FBO': [
            'site_templates/*.py',
            'site_templates/requirements.txt',
            'site_templates/static/css/*.css',
            'site_templates/templates/*.html',
            'site_templates/templates/blog/*.html',
        ],
    },
    license='MIT',
    author='James Aylett',
    author_email='james@tartarus.org',
    entry_points={
        'console_scripts': [
            'django-fbo-newsite = django_FBO.__script__:newsite'
        ],
    },
    install_requires=[
        'Django~=1.10',
        'PyYAML~=3.11',
    ],
    url = 'https://github.com/jaylett/django-filebacked-objects',
    classifiers = [
        'Intended Audience :: Developers',
        'Framework :: Django',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)



import collections
from operator import attrgetter
from django.conf import settings
from django.contrib.staticfiles import utils
from django.core.files.storage import FileSystemStorage
from django.utils import timezone

from .file_objects import FileObject
from .query import Q


OPTS = [
    'path',
    'metadata',
    'model',
    'slug_suffices',
    'slug_strip_index',
    'storage',
    '_filters',
    '_order_by',
    '_slice',
    '_fetched',
]


class FBO:
    storage = FileSystemStorage
    path = None
    metadata = None
    glob = None
    slug_suffices = None
    slug_strip_index = False
    model = FileObject

    _filters = None
    _order_by = None
    _slice = None

    def __init__(self, **kwargs):
        self._fetched = None
        if self._filters is None:
            self._filters = []
        else:
            self._filters = self._filters[:]
        if self._order_by is None:
            self._order_by = []
        else:
            self._order_by = self._order_by[:]

        for opt in OPTS:
            if opt in kwargs:
                if kwargs[opt] is not None:
                    setattr(self, opt, kwargs[opt])
        if 'glob' in kwargs and kwargs['glob'] is not None:
            self._add_q(
                Q(name__glob=kwargs['glob']),
            )
        elif self.glob is not None:
            self._add_q(
                Q(name__glob=self.glob),
            )
        
        if self.path is None:
            raise TypeError("You must set a path for FBOs.")

        self._storage = self.storage(
            location=self.path,
        )

    def clone(self, **kwargs):
        # subclass this if your subclass has more attributes
        # rather than just changing defaults
        for opt in OPTS:
            val = getattr(self, opt)
            if val is not None:
                if opt == '_slice':
                    if kwargs['_slice'] is None:
                        kwargs['_slice'].setdfeault(val)
                    else:
                        # merge with existing slice
                        def _combine_start(one, two):
                            if one.start is None:
                                s1 = 0
                            else:
                                s1 = one.start
                            if two.start is None:
                                s2 = 0
                            else:
                                s2 = two.start
                            return s1 + s2

                        def _combine_stop(one, two):
                            # stop is an index, not an offset
                            # so we have to take start into account
                            if one.stop is None and two.stop is None:
                                return None
                            elif two.stop is None:
                                # it's one.stop, adjusted for start
                                # change if there was one
                                if two.start is None:
                                    return one.stop
                                else:
                                    return one.stop + two.start
                            elif one.stop is None:
                                # it's two.stop, adjusted as above
                                if one.start is None:
                                    return two.stop
                                else:
                                    return two.stop + one.start
                            else:
                                # both have stop, so adjust and take
                                # the minimum
                                if two.start is None:
                                    s1 = one.stop
                                else:
                                    s1 = one.stop + two.start
                                if one.start is None:
                                    s2 = two.stop
                                else:
                                    s2 = two.stop + one.start
                                return min(s1, s2)

                        def _combine_step(one, two):
                            if one.step is None:
                                s1 = 1
                            else:
                                s1 = one.step
                            if two.step is None:
                                s2 = 1
                            else:
                                s2 = two.step
                            return s1 * s2

                        #print(val, 'merge', kwargs['_slice'])
                        kwargs['_slice'] = slice(
                            _combine_start(val, kwargs['_slice']),
                            _combine_stop(val, kwargs['_slice']),
                            _combine_step(val, kwargs['_slice']),
                        )
                        #print('now', kwargs['_slice'])
                # Lists must be duplicated, not just copied by
                # reference, otherwise if we mutate them in-place
                # in the clone it will affect the parent. This is
                # particularly bad if you reuse FBO objects, which
                # you probably will for instance in generic CBVs.
                elif isinstance(val, collections.Iterable):
                    kwargs.setdefault(opt, val[:])
                else:
                    kwargs.setdefault(opt, val)
        # cached data
        kwargs['_fetched'] = self._fetched
        return type(self)(**kwargs)

    @property
    def objects(self):
        return self

    def all(self):
        return self.clone()

    def none(self):
        return self.filter(name=False)

    def exists(self):
        return self.count() > 0

    def datetimes(self, field_name, kind, ordering, tzinfo=None):
        if settings.USE_TZ:
            if tzinfo is None:
                tzinfo = timezone.get_current_timezone()
        else:
            tzinfo = None

        datetimes = set()
        for obj in iter(self):
            dt = getattr(obj, field_name)
            reset = {
                'month': 1,
                'day': 1,
                'hour': 0,
                'minute': 0,
                'second': 0,
                'microsecond': 0,
            }
            for _kind in 'year', 'month', 'day', 'hour', 'minute', 'second':
                if _kind in reset.keys():
                    del reset[_kind]
                if _kind == kind:
                    break
            datetimes.add(dt.replace(**reset))

        if ordering == 'ASC':
            return sorted(datetimes)
        else:
            return sorted(datetimes, reverse=True)

    def _add_q(self, q_object):
        self._filters.append(q_object)
    
    def filter(self, *args, **kwargs):
        clone = self.clone()
        clone._add_q(Q(*args, **kwargs))
        return clone
    
    def exclude(self, *args, **kwargs):
        clone = self.clone()
        clone._add_q(~Q(*args, **kwargs))
        return clone
    
    def __len__(self):
        return self.count()
    
    def __getitem__(self, idx):
        if isinstance(idx, slice):
            if (
                idx.start is not None and idx.start < 0
            ) or (
                idx.stop is not None and idx.stop < 0
            ):
                raise ValueError("Cannot use negative indexes with FBO.")
            return self.clone(_slice=idx)
        _iter = iter(self)
        try:
            for i in range(0, idx+1):
                obj = next(_iter)
        except StopIteration:
            raise IndexError(idx)
        return obj
    
    def order_by(self, *args):
        return self.clone(_order_by=args)

    def get(self, *args, **kwargs):
        filtered = self.clone().filter(*args, **kwargs)
        _count = filtered.count()
        if _count == 0:
            raise self.model.DoesNotExist
        elif _count > 1:
            raise self.model.MultipleObjectsReturned
        else:
            return filtered[0]
    
    def count(self):
        _count = 0
        for _file in iter(self):
            _count += 1
        return _count

    def _prefetch(self):
        if self._fetched is None or settings.DEBUG:
            self._fetched = []
            #print("Starting _prefetch")
            for fname in utils.get_files(self._storage):
                #print("Found %s." % fname)
                _file = self.model(
                    self._storage,
                    self.metadata,
                    fname,
                    self.slug_suffices,
                    self.slug_strip_index,
                )
                if self._check_filters(_file):
                    #print("  matches filters")
                    self._fetched.append(_file)
    
    def __iter__(self):
        self._prefetch()
        # apply order_by here because we may have prefetched on a
        # previous copy of this
        _objects = self._fetched
        for _order_by in self._order_by:
            if _order_by[0] == '-':
                _rev = True
                _attr = _order_by[1:]
            else:
                _rev = False
                _attr = _order_by
            _objects = sorted(_objects, key=attrgetter(_attr), reverse=_rev)

        # check filters here as well as so we can copy
        # our cached fetched data around, to avoid hitting
        # the filesystem so much
        #print("__iter__()")
        _filtered = []
        for _file in _objects:
            #print("Considering %s" % _file)
            if self._check_filters(_file):
                #print("  matches filters.")
                _filtered.append(_file)
        if self._slice is not None:
            #print('sliced', self._slice, [o.name for o in _filtered])
            _filtered = _filtered.__getitem__(self._slice)
            #print('now', [o.name for o in _filtered])
        return iter(_filtered)

    def _check_filters(self, _file):
        for _filter in self._filters:
            if not _filter(_file):
                return False
        return True
